/* Syscall function set for newlib */
/* Copyright (C) Shinichi Yamagiwa */

#include <reent.h>
#include <unistd.h>
#include <sys/stat.h>
#include <_syslist.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/times.h>
#include <sys/time.h>

#include "target_syscalls.h"

/*

3) Define or otherwise provide the regular versions of the syscalls
   (eg: open, close, etc.).  The library won't be reentrant nor
   namespace clean, but at least it will work.  When you do this, add
   -DMISSING_SYSCALL_NAMES to newlib_cflags (TARGET_CFLAGS) in
   configure.host.

   add --disable-newlib-supplied-syscalls to configure

   The following functions listed in newlib/libc/include/_syslist.h
   must be ported to target environment;

   int _close (int fd);
   int _execve (char *name, char **argv, char **env);
   int _fcntl (int fd, int flag, int arg);
   int _fork ();
   int _fstat (int fd, struct stat *pstat);
   int _getpid ();
   int _gettimeofday (struct timeval *ptimeval, struct timezone *ptimezone);
   int _kill (int pid, int sig);
   int _link (char *old, char *new);
   int _lseek (int fd, off_t pos, int whence);
   int _open (const char *file, int flags);
   int _read (int fd, void *buf, size_t cnt);
   int _sbrk (ptrdiff_t incr);
   int _stat (char *file, struct stat *pstat);
   int _times (struct tms *buf);
   int _unlink (char *file);
   int _wait (int *status);
   int _write (int fd, const void *buf, size_t cnt);

*/

#include <errno.h>
#undef errno
extern int errno;

/* A pointer to a list of environment variables and their values. 
   For a minimal environment, this empty list is adequate. */
char *__env[1] = { 0 };
char **environ = __env;


/* =================== File System Related Functions ================ */
/* Close a file. Minimal implementation: */
int _close (int fd){
  return -1;
}

/* Establish a new name for an existing file. Minimal implementation: */
int _link (const char *old, const char *new){
  errno = EMLINK;
  return -1;
}

// File
/* Set position in a file. Minimal implementation: */
_off_t _lseek (int fd, off_t pos, int whence){
  return 0;
}

// File
/* Open a file. Minimal implementation: */
int _open (const char *file, int flags){
  return -1;
}

/* Status of a file (by name). Minimal implementation: */
int _stat (char *file, struct stat *pstat){
  pstat->st_mode = S_IFCHR;
  return 0;
}

/* Remove a file's directory entry. Minimal implementation: */
int _unlink (char *file){
  errno = ENOENT;
  return -1;
}


/* =================== Process Control Related Functions ============ */
/* Transfer control to a new process. Minimal implementation (for a
   system without processes): */
int _execve (const char *name, char * const argv[], char * const env[]){
  errno = ENOMEM;
  return -1;
}

int _fcntl (int fd, int flag, int arg){
  errno = EAGAIN;
  return -1;
}

/* Create a new process. Minimal implementation (for a system without
   processes): */
int _fork (){
  errno = EAGAIN;
  return -1;
}

/* Status of an open file. For consistency with other minimal
   implementations in these examples, all files are regarded as
   character special devices. The sys/stat.h header file required is
   distributed in the include subdirectory for this C library. */
int _fstat (int fd, struct stat *pstat){
  pstat->st_mode = S_IFCHR;
  return 0;
}

/* Process-ID; this is sometimes used to generate strings unlikely to
   conflict with other processes. Minimal implementation, for a system
   without processes: */
int _getpid (){
  return 1;
}

/* Wait for a child process. Minimal implementation: */
int _wait (int *status){
  errno = ECHILD;
  return -1;
}

/* Query whether output stream is a terminal. For consistency with the
   other minimal implementations, which only support output to stdout,
   this minimal implementation is suggested:*/
int isatty (int fd){
  return (fd <= 2) ? 1 : 0;  /* one of stdin, stdout, stderr */
}

/* Send a signal. Minimal implementation: */
int _kill (int pid, int sig){
  errno = EINVAL;
  return -1;
}



/* ===========  File I/O Related Functions ============= */
/* Read from a file. Minimal implementation: */
int _read (int fd, void *buf, size_t cnt){
  int buf_len;
  if(fd == 0){ // stdin
    if(__StdInIsInitialized() == 0){ // stdin is not initialized.
      if(__InitializeStdIn() == 0) return 0;
    }
    buf_len = cnt;
    buf_len = __InputFromStdIn((unsigned char *)buf, buf_len);
    return (size_t)buf_len;
  }
  return 0;
}

/* Write to a file. libc subroutines will use this system routine for
   output to all files, including stdout?so if you need to generate
   any output, for example to a serial port for debugging, you should
   make your minimal write capable of doing this. The following
   minimal implementation is an incomplete example; it relies on a
   outbyte subroutine (not shown; typically, you must write this in
   assembler from examples provided by your hardware manufacturer) to
   actually perform the output. */

int _write (int fd, const void *buf, size_t cnt){
  // output to UART
  int buf_length;

  if((fd  == 1) || (fd == 2)){ // stdout or stderr
    if(__StdOutIsInitialized() == 0){ // stdout is not initialized.
      if(__InitializeStdOut() == 0) // Failed to initialize
	return 0;
    }
    buf_length = (int)cnt;
    buf_length = __OutputToStdOut((unsigned char *)buf, buf_length);
    return (size_t)buf_length;
  }
  return 0;
}

/* ============= Memory Allocation Related Functions ============= */

/* Increase program data space. As malloc and related functions depend
   on this, it is useful to have a working implementation. The
   following suffices for a standalone system; it exploits the symbol
   _end automatically defined by the GNU linker. */

extern char __heap_top;		/* Defined by the linker */
extern char __heap_end;		/* Defined by the linker */
char *curr_heap_end = 0;


caddr_t _sbrk (ptrdiff_t incr){
  char *prev_heap_end;
  
  if (curr_heap_end == 0) {
    curr_heap_end = &__heap_top;
  }
  prev_heap_end = curr_heap_end;
  if ((curr_heap_end + incr) > &__heap_end) {
    write (2, "Heap and stack collision. Stoped.\n\r", 26);
    while(1){}
  }
  
  curr_heap_end += incr;
  return (caddr_t) prev_heap_end;
}



/* ========== Time Related Functions ============== */
int _gettimeofday (struct timeval *ptimeval, struct timezone *ptimezone){
  // ignore timezone struct.

  if(__TimerIsActive() == 0){ // Timer is not active.
    if(__ActivateTimer() == 0) // Failed to initialize timer.
      return 0;
    int sec;
    int usec;
    if(__GetCurrTime(&sec, &usec) == 0)  return -1;
    ptimeval->tv_sec = (time_t)sec;
    ptimeval->tv_usec = (suseconds_t)usec;
  }
  return 0;
}

/* Timing information for current process. Minimal implementation: */
int _times (struct tms *buf){
  return -1;
}

/* ===================== Abort Related Functions ============= */
void _exit(int ret){
  __AbnormalExit();
}
